/*
  UTFT.cpp - Multi-Platform library support for Color TFT LCD Boards
  Copyright (C)2015 Rinky-Dink Electronics, Henning Karlsen. All right reserved
  
  This library is the continuation of my ITDB02_Graph, ITDB02_Graph16
  and RGB_GLCD libraries for Arduino and chipKit. As the number of 
  supported display modules and controllers started to increase I felt 
  it was time to make a single, universal library as it will be much 
  easier to maintain in the future.

  Basic functionality of this library was origianlly based on the 
  demo-code provided by ITead studio (for the ITDB02 modules) and 
  NKC Electronics (for the RGB GLCD module/shield).

  This library supports a number of 8bit, 16bit and serial graphic 
  displays, and will work with both Arduino, chipKit boards and select 
  TI LaunchPads. For a full list of tested display modules and controllers,
  see the document UTFT_Supported_display_modules_&_controllers.pdf.

  When using 8bit and 16bit display modules there are some 
  requirements you must adhere to. These requirements can be found 
  in the document UTFT_Requirements.pdf.
  There are no special requirements when using serial displays.

  You can find the latest version of the library at 
  http://www.RinkyDinkElectronics.com/

  This library is free software; you can redistribute it and/or
  modify it under the terms of the CC BY-NC-SA 3.0 license.
  Please see the included documents for further information.

  Commercial use of this library requires you to buy a license that
  will allow commercial use. This includes using the library,
  modified or not, as a tool to sell products.

  The license applies to all part of the library including the 
  examples and tools supplied with the library.
*/


#include <math.h>
#include <stdio.h>
#include <string.h>

#include "UTFT.h"

//#include "config/default/peripheral/gpio/plib_gpio.h"

//#include "memorysaver.h"

struct _current_font
{
	uint8_t* font;
	uint8_t x_size;
	uint8_t y_size;
	uint8_t offset;
	uint8_t numchars;
}cfont;


uint8_t			fch, fcl, bch, bcl;
uint8_t			orient;
long			disp_x_size, disp_y_size;
uint8_t			display_model, display_transfer_mode, display_serial_mode;
bool			_transparent;
        
void _convert_float(char *buf, double num, int width, uint8_t prec)
{
	char format[10];
	
	sprintf(format, "%%%i.%if", width, prec);
	sprintf((char *)buf, format, num);
}

void _fast_fill_8(int ch, long pix)
{
	long blocks;
    int i;

	LATECLR = 0xFF;
	LATESET = ch;

	blocks = pix/16;
    
	for (i=0; i<blocks*16; i++)
        LCD_Write_BUS(ch, ch);

	if ((pix % 16) != 0){
		for (i=0; i<(pix % 16)+1; i++)
		LCD_Write_BUS(ch,  ch);
	}	
}

void LCD_Write_COM(uint8_t VL)  
{      
    //PLIB_PMP_AddressSet(PMP_ID_0, 0);
    //DRV_PMP0_Write(VL);
    PMP_AddressSet(0);
    PMP_MasterSend(VL);
}

void LCD_Write_BUS(uint8_t VH, uint8_t VL)
{
    /*
    PLIB_PMP_AddressSet(PMP_ID_0, 1);
    DRV_PMP0_Write(VH);

    PLIB_PMP_AddressSet(PMP_ID_0, 1);
    DRV_PMP0_Write(VL);
    */
    
    PMP_AddressSet(1);
    PMP_MasterSend(VH);
    
    PMP_AddressSet(1);
    PMP_MasterSend(VL);
}

void LCD_Write_DATA(uint8_t VL)
{
//    PLIB_PMP_AddressSet(PMP_ID_0, 1);
//    DRV_PMP0_Write(VL); 
    
    PMP_AddressSet(1);
    PMP_MasterSend(VL);  
    
}

void LCD_Write_COM_DATA(char com1,int dat1)
{
     LCD_Write_COM(com1);
     LCD_Write_BUS(dat1>>8,dat1);
       
}

void InitLCD(uint8_t orientation)
{
	orient=orientation;
    
	disp_x_size =			239;
	disp_y_size =			319;
	display_transfer_mode =	8;//dtm[model];
	display_model =			1;//model;
    
	RST_ON;
	DelayMs(1); 
	RST_OFF;
	DelayMs(10);
	RST_ON; 
	DelayMs(50);
/*
	sbi(P_RST, B_RST);
	delay(5); 
	cbi(P_RST, B_RST);
	delay(15);
	sbi(P_RST, B_RST);
	delay(15);
*/
    
	CS_EN;//cbi(P_CS, B_CS);

   //case ILI9341_16:
    LCD_Write_COM(0xCB);  
    LCD_Write_DATA(0x39); 
    LCD_Write_DATA(0x2C); 
    LCD_Write_DATA(0x00); 
    LCD_Write_DATA(0x34); 
    LCD_Write_DATA(0x02); 

    LCD_Write_COM(0xCF);  
    LCD_Write_DATA(0x00); 
    LCD_Write_DATA(0XC1); 
    LCD_Write_DATA(0X30); 

    LCD_Write_COM(0xE8);  
    LCD_Write_DATA(0x85); 
    LCD_Write_DATA(0x00); 
    LCD_Write_DATA(0x78); 

    LCD_Write_COM(0xEA);  
    LCD_Write_DATA(0x00); 
    LCD_Write_DATA(0x00); 
 
    LCD_Write_COM(0xED);  
    LCD_Write_DATA(0x64); 
    LCD_Write_DATA(0x03); 
    LCD_Write_DATA(0X12); 
    LCD_Write_DATA(0X81); 

    LCD_Write_COM(0xF7);  
    LCD_Write_DATA(0x20); 
  
    LCD_Write_COM(0xC0);    //Power control 
    LCD_Write_DATA(0x23);   //VRH[5:0] 
 
    LCD_Write_COM(0xC1);    //Power control 
    LCD_Write_DATA(0x10);   //SAP[2:0];BT[3:0] 

    LCD_Write_COM(0xC5);    //VCM control 
    LCD_Write_DATA(0x3e);   //Contrast
    LCD_Write_DATA(0x28); 
 
    LCD_Write_COM(0xC7);    //VCM control2 
    LCD_Write_DATA(0x86);   //--
 
    LCD_Write_COM(0x36);    // Memory Access Control 
    LCD_Write_DATA(0x48);   

    LCD_Write_COM(0x3A);    
    LCD_Write_DATA(0x55); 

    LCD_Write_COM(0xB1);    
    LCD_Write_DATA(0x00);  
    LCD_Write_DATA(0x18); 
 
    LCD_Write_COM(0xB6);    // Display Function Control 
    LCD_Write_DATA(0x08); 
    LCD_Write_DATA(0x82);
    LCD_Write_DATA(0x27);  
/* 
    LCD_Write_COM(0xF2);    // 3Gamma Function Disable 
    LCD_Write_DATA(0x00); 
 
    LCD_Write_COM(0x26);    //Gamma curve selected 
    LCD_Write_DATA(0x01); 

    LCD_Write_COM(0xE0);    //Set Gamma 
    LCD_Write_DATA(0x0F); 
    LCD_Write_DATA(0x31); 
    LCD_Write_DATA(0x2B); 
    LCD_Write_DATA(0x0C); 
    LCD_Write_DATA(0x0E); 
    LCD_Write_DATA(0x08); 
    LCD_Write_DATA(0x4E); 
    LCD_Write_DATA(0xF1); 
    LCD_Write_DATA(0x37); 
    LCD_Write_DATA(0x07); 
    LCD_Write_DATA(0x10); 
    LCD_Write_DATA(0x03); 
    LCD_Write_DATA(0x0E); 
    LCD_Write_DATA(0x09); 
    LCD_Write_DATA(0x00); 

    LCD_Write_COM(0XE1);    //Set Gamma 
    LCD_Write_DATA(0x00); 
    LCD_Write_DATA(0x0E); 
    LCD_Write_DATA(0x14); 
    LCD_Write_DATA(0x03); 
    LCD_Write_DATA(0x11); 
    LCD_Write_DATA(0x07); 
    LCD_Write_DATA(0x31); 
    LCD_Write_DATA(0xC1); 
    LCD_Write_DATA(0x48); 
    LCD_Write_DATA(0x08); 
    LCD_Write_DATA(0x0F); 
    LCD_Write_DATA(0x0C); 
    LCD_Write_DATA(0x31); 
    LCD_Write_DATA(0x36); 
    LCD_Write_DATA(0x0F); 
*/
    LCD_Write_COM(0x11);    //Exit Sleep 
    	DelayMs(120); 
				
    LCD_Write_COM(0x29);    //Display on 
    LCD_Write_COM(0x2c); 

	CS_DISEN;//sbi (P_CS, B_CS); 

	cfont.font=0;
	_transparent = false;
}

void setXY(uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2)
{
	if (orient==LANDSCAPE)
	{
		swap(uint16_t, x1, y1);
		swap(uint16_t, x2, y2)
		y1=disp_y_size-y1;
		y2=disp_y_size-y2;
		swap(uint16_t, y1, y2)
	}

	LCD_Write_COM(0x2a);
  	LCD_Write_DATA(x1>>8);
  	LCD_Write_DATA(x1);
  	LCD_Write_DATA(x2>>8);
  	LCD_Write_DATA(x2);
  	LCD_Write_COM(0x2b);
  	LCD_Write_DATA(y1>>8);
  	LCD_Write_DATA(y1);
  	LCD_Write_DATA(y2>>8);
  	LCD_Write_DATA(y2);
  	LCD_Write_COM(0x2c); 
}

void getXY(uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2)
{
	if (orient==LANDSCAPE)
	{
		swap(uint16_t, x1, y1);
		swap(uint16_t, x2, y2)
		y1=disp_y_size-y1;
		y2=disp_y_size-y2;
		swap(uint16_t, y1, y2)
	}

	LCD_Write_COM(0x2a);
  	LCD_Write_DATA(x1>>8);
  	LCD_Write_DATA(x1);
  	LCD_Write_DATA(x2>>8);
  	LCD_Write_DATA(x2);
  	LCD_Write_COM(0x2b);
  	LCD_Write_DATA(y1>>8);
  	LCD_Write_DATA(y1);
  	LCD_Write_DATA(y2>>8);
  	LCD_Write_DATA(y2);
  	LCD_Write_COM(0x2e); 
}

void clrXY()
{
	if (orient==PORTRAIT)
		setXY(0,0,disp_x_size,disp_y_size);
	else
		setXY(0,0,disp_y_size,disp_x_size);
}

void drawRect(int x1, int y1, int x2, int y2)
{
	if (x1>x2)
	{
		swap(int, x1, x2);
	}
	if (y1>y2)
	{
		swap(int, y1, y2);
	}

	drawHLine(x1, y1, x2-x1);
	drawHLine(x1, y2, x2-x1);
	drawVLine(x1, y1, y2-y1);
	drawVLine(x2, y1, y2-y1);
}

void drawRoundRect(int x1, int y1, int x2, int y2)
{
	if (x1>x2)
	{
		swap(int, x1, x2);
	}
	if (y1>y2)
	{
		swap(int, y1, y2);
	}
	if ((x2-x1)>4 && (y2-y1)>4)
	{
		drawPixel(x1+1,y1+1);
		drawPixel(x2-1,y1+1);
		drawPixel(x1+1,y2-1);
		drawPixel(x2-1,y2-1);
		drawHLine(x1+2, y1, x2-x1-4);
		drawHLine(x1+2, y2, x2-x1-4);
		drawVLine(x1, y1+2, y2-y1-4);
		drawVLine(x2, y1+2, y2-y1-4);
	}
}

void fillRect(int x1, int y1, int x2, int y2)
{
    int i;
    
	if (x1>x2)
	{
		swap(int, x1, x2);
	}
	if (y1>y2)
	{
		swap(int, y1, y2);
	}

    if (fch==fcl)
	{
		CS_EN;//cbi(P_CS, B_CS);
		setXY(x1, y1, x2, y2);
//		RS_PARA;//sbi(P_RS, B_RS);
		_fast_fill_8(fch,(((long)(x2-x1)+1)*((long)(y2-y1)+1)));
		CS_DISEN;//sbi(P_CS, B_CS);
	}
	else
	{
		if (orient==PORTRAIT)
		{
			for (i=0; i<((y2-y1)/2)+1; i++)
			{
				drawHLine(x1, y1+i, x2-x1);
				drawHLine(x1, y2-i, x2-x1);
			}
		}
		else
		{
			for (i=0; i<((x2-x1)/2)+1; i++)
			{
				drawVLine(x1+i, y1, y2-y1);
				drawVLine(x2-i, y1, y2-y1);
			}
		}
	}
}

void fillRoundRect(int x1, int y1, int x2, int y2)
{
    int i;
    
	if (x1>x2)
	{
		swap(int, x1, x2);
	}
	if (y1>y2)
	{
		swap(int, y1, y2);
	}

	if ((x2-x1)>4 && (y2-y1)>4)
	{
		for (i=0; i<((y2-y1)/2)+1; i++)
		{
			switch(i)
			{
			case 0:
				drawHLine(x1+2, y1+i, x2-x1-4);
				drawHLine(x1+2, y2-i, x2-x1-4);
				break;
			case 1:
				drawHLine(x1+1, y1+i, x2-x1-2);
				drawHLine(x1+1, y2-i, x2-x1-2);
				break;
			default:
				drawHLine(x1, y1+i, x2-x1);
				drawHLine(x1, y2-i, x2-x1);
			}
		}
	}
}

void drawCircle(int x, int y, int radius)
{
	int f = 1 - radius;
	int ddF_x = 1;
	int ddF_y = -2 * radius;
	int x1 = 0;
	int y1 = radius;
 
	CS_EN;//cbi(P_CS, B_CS);
	setXY(x, y + radius, x, y + radius);
	LCD_Write_BUS(fch,fcl);
	setXY(x, y - radius, x, y - radius);
	LCD_Write_BUS(fch,fcl);
	setXY(x + radius, y, x + radius, y);
	LCD_Write_BUS(fch,fcl);
	setXY(x - radius, y, x - radius, y);
	LCD_Write_BUS(fch,fcl);
 
	while(x1 < y1)
	{
		if(f >= 0) 
		{
			y1--;
			ddF_y += 2;
			f += ddF_y;
		}
		x1++;
		ddF_x += 2;
		f += ddF_x;    
		setXY(x + x1, y + y1, x + x1, y + y1);
		LCD_Write_BUS(fch,fcl);
		setXY(x - x1, y + y1, x - x1, y + y1);
		LCD_Write_BUS(fch,fcl);
		setXY(x + x1, y - y1, x + x1, y - y1);
		LCD_Write_BUS(fch,fcl);
		setXY(x - x1, y - y1, x - x1, y - y1);
		LCD_Write_BUS(fch,fcl);
		setXY(x + y1, y + x1, x + y1, y + x1);
		LCD_Write_BUS(fch,fcl);
		setXY(x - y1, y + x1, x - y1, y + x1);
		LCD_Write_BUS(fch,fcl);
		setXY(x + y1, y - x1, x + y1, y - x1);
		LCD_Write_BUS(fch,fcl);
		setXY(x - y1, y - x1, x - y1, y - x1);
		LCD_Write_BUS(fch,fcl);
	}
	CS_DISEN;//sbi(P_CS, B_CS);
	clrXY();
}

void fillCircle(int x, int y, int radius)
{
    int x1, y1;
    
	for(y1=-radius; y1<=0; y1++) 
		for(x1=-radius; x1<=0; x1++)
			if(x1*x1+y1*y1 <= radius*radius) 
			{
				drawHLine(x+x1, y+y1, 2*(-x1));
				drawHLine(x+x1, y-y1, 2*(-x1));
				break;
			}
}

void show_color_bar()
{
	unsigned long i,j;
	CS_EN;
	clrXY();

		for (i=0; i<(disp_y_size+1); i++)
			{
				for (j=0; j<(disp_x_size+1); j++)
				{			
	 				if(i>350) LCD_Write_BUS(0xFF, 0xFF);
 	 				else if(i>300) LCD_Write_BUS(0x00, 0x1F);
 	 				else if(i>250) LCD_Write_BUS(0x07, 0xE0);
 	 				else if(i>200) LCD_Write_BUS(0x7F, 0xFF);
 	 				else if(i>150) LCD_Write_BUS(0xF8, 0x00);
 	 				else if(i>100) LCD_Write_BUS(0xF8, 0x1F);
 	 				else if(i>50) LCD_Write_BUS(0xFF, 0xE0);
					else LCD_Write_BUS(213, 156);
				}	
			}
    CS_DISEN;
}

void clrScr()
{
	
	CS_EN;//cbi(P_CS, B_CS);
	clrXY();

//	RS_PARA;//sbi(P_RS, B_RS);
	_fast_fill_8(0,((disp_x_size+1)*(disp_y_size+1)));
	
	CS_DISEN;//sbi(P_CS, B_CS);
}

void fillScr(uint8_t r, uint8_t g, uint8_t b)
{
	uint16_t color = ((r&248)<<8 | (g&252)<<3 | (b&248)>>3);
	fillScr_w(color);
}

void fillScr_w(uint16_t color)
{
	long i;
	char ch, cl;
	
	ch=(uint8_t)(color>>8);
	cl=(uint8_t)(color & 0xFF);

	CS_EN;//cbi(P_CS, B_CS);
	clrXY();
    
    if (ch==cl){
//        RS_PARA;//sbi(P_RS, B_RS);
		_fast_fill_8(ch,((disp_x_size+1)*(disp_y_size+1)));
    }
	else
	{
		for (i=0; i<((disp_x_size+1)*(disp_y_size+1)); i++)
		{	
				LCD_Write_BUS(ch,cl);
		}
	}
	CS_DISEN;//sbi(P_CS, B_CS);
}

void setColor(uint8_t r, uint8_t g, uint8_t b)
{
	fch=((r&248)|g>>5);
	fcl=((g&28)<<3|b>>3);
}

void setColor_w(uint16_t color)
{
	fch=(uint8_t)(color>>8);
	fcl=(uint8_t)(color & 0xFF);
}

uint16_t getColor()
{
	return (fch<<8) | fcl;
}

void setBackColor(uint8_t r, uint8_t g, uint8_t b)
{
	bch=((r&248)|g>>5);
	bcl=((g&28)<<3|b>>3);
	_transparent=false;
}

void setBackColor_w(uint32_t color)
{
	if (color==VGA_TRANSPARENT)
		_transparent=true;
	else
	{
		bch=(uint8_t)(color>>8);
		bcl=(uint8_t)(color & 0xFF);
		_transparent=false;
	}
}

uint16_t getBackColor()
{
	return (bch<<8) | bcl;
}

void setPixel(uint16_t color)
{
	LCD_Write_BUS((color>>8),(color&0xFF));	// rrrrrggggggbbbbb
}

uint16_t getPixel(int x, int y){
    uint16_t VH,VM,VL;
    uint16_t i;

	CS_EN;
	getXY(x, y, x, y);
    
//    SYS_PORTS_DirectionSelect(  PORTS_ID_0, SYS_PORTS_DIRECTION_INPUT, PORT_CHANNEL_E, 0xff);

    GPIO_PortInputEnable(GPIO_PORT_E, 0xff);
 
    //dummy read 16 bits
//    PLIB_PMP_AddressSet(PMP_ID_0, 1);
//    VH = DRV_PMP0_Read();

    PMP_AddressSet(1);
    VH= PMP_MasterReceive();
    
//    PLIB_PMP_AddressSet(PMP_ID_0, 1);
//    VH = DRV_PMP0_Read();

    PMP_AddressSet(1);
    VH= PMP_MasterReceive();
    
    //real read 16 bits
//    PLIB_PMP_AddressSet(PMP_ID_0, 1);
//    VH = DRV_PMP0_Read();

    PMP_AddressSet(1);
    VH= PMP_MasterReceive();

   
//    PLIB_PMP_AddressSet(PMP_ID_0, 1);
//    VM = DRV_PMP0_Read();

    PMP_AddressSet(1);
    VM= PMP_MasterReceive();
    
//    PLIB_PMP_AddressSet(PMP_ID_0, 1);
//    VL = DRV_PMP0_Read();

    PMP_AddressSet(1);
    VL= PMP_MasterReceive();
    
//    SYS_PORTS_DirectionSelect(  PORTS_ID_0, SYS_PORTS_DIRECTION_OUTPUT, PORT_CHANNEL_E, 0xff);
        GPIO_PortOutputEnable(GPIO_PORT_E, 0xff);
        
    i=(VH & 0xf8)<<8;
    i+=(VM & 0xfc) <<3;
    i+=(VL & 0xf8)>>3;

    CS_DISEN;
	clrXY();


    return i;
}

void drawPixel(int x, int y)
{
	CS_EN;
	setXY(x, y, x, y);
	setPixel((fch<<8)|fcl);
	CS_DISEN;
	clrXY();
}

void drawLine(int x1, int y1, int x2, int y2)
{
	if (y1==y2)
		drawHLine(x1, y1, x2-x1);
	else if (x1==x2)
		drawVLine(x1, y1, y2-y1);
	else
	{
		unsigned int	dx = (x2 > x1 ? x2 - x1 : x1 - x2);
		short			xstep =  x2 > x1 ? 1 : -1;
		unsigned int	dy = (y2 > y1 ? y2 - y1 : y1 - y2);
		short			ystep =  y2 > y1 ? 1 : -1;
		int				col = x1, row = y1;

		CS_EN;//cbi(P_CS, B_CS);
		if (dx < dy)
		{
			int t = - (dy >> 1);
			while (true)
			{
				setXY (col, row, col, row);
				LCD_Write_BUS(fch, fcl);
				if (row == y2)
					return;
				row += ystep;
				t += dx;
				if (t >= 0)
				{
					col += xstep;
					t   -= dy;
				}
			} 
		}
		else
		{
			int t = - (dx >> 1);
			while (true)
			{
				setXY (col, row, col, row);
				LCD_Write_BUS(fch, fcl);
				if (col == x2)
					return;
				col += xstep;
				t += dy;
				if (t >= 0)
				{
					row += ystep;
					t   -= dx;
				}
			} 
		}
		CS_DISEN;//sbi(P_CS, B_CS);
	}
	clrXY();
}

void drawHLine(int x, int y, int l)
{
    int i;
    
	if (l<0)
	{
		l = -l;
		x -= l;
	}
	CS_EN;//cbi(P_CS, B_CS);
	setXY(x, y, x+l, y);

    if (fch==fcl)
	{
//		RS_PARA;//sbi(P_RS, B_RS);
		_fast_fill_8(fch,l);
	}
	else
	{
		for (i=0; i<l+1; i++)
		{
			LCD_Write_BUS(fch, fcl);
		}
	}
	CS_DISEN;//sbi(P_CS, B_CS);
	clrXY();
}

void drawVLine(int x, int y, int l)
{
    int i;
    
	if (l<0)
	{
		l = -l;
		y -= l;
	}
	CS_EN;//cbi(P_CS, B_CS);
	setXY(x, y, x, y+l);

    if (fch==fcl)
	{
//		RS_PARA;//sbi(P_RS, B_RS);
		_fast_fill_8(fch,l);
	}
	else
	{
		for (i=0; i<l+1; i++)
		{
			LCD_Write_BUS(fch, fcl);
		}
	}
	CS_DISEN;//sbi(P_CS, B_CS);
	clrXY();
}

void printChar(uint8_t c, int x, int y)
{
	uint8_t i,ch;
	uint16_t j;
	uint16_t temp;
    int zz;

	CS_EN;//cbi(P_CS, B_CS);
  
	if (!_transparent)
	{
		if (orient==PORTRAIT)
		{
			setXY(x,y,x+cfont.x_size-1,y+cfont.y_size-1);
	  
			temp=((c-cfont.offset)*((cfont.x_size/8)*cfont.y_size))+4;
			for(j=0;j<((cfont.x_size/8)*cfont.y_size);j++)
			{
				ch=cfont.font[temp];//ch=pgm_read_byte(&cfont.font[temp]);
				for(i=0;i<8;i++)
				{   
					if((ch&(1<<(7-i)))!=0)   
					{
						setPixel((fch<<8)|fcl);
					} 
					else
					{
						setPixel((bch<<8)|bcl);
					}   
				}
				temp++;
			}
		}
		else
		{
			temp=((c-cfont.offset)*((cfont.x_size/8)*cfont.y_size))+4;

			for(j=0;j<((cfont.x_size/8)*cfont.y_size);j+=(cfont.x_size/8))
			{
				setXY(x,y+(j/(cfont.x_size/8)),x+cfont.x_size-1,y+(j/(cfont.x_size/8)));
				for (zz=(cfont.x_size/8)-1; zz>=0; zz--)
				{
					ch=cfont.font[temp+zz];//ch=pgm_read_byte(&cfont.font[temp+zz]);
					for(i=0;i<8;i++)
					{   
						if((ch&(1<<i))!=0)   
						{
							setPixel((fch<<8)|fcl);
						} 
						else
						{
							setPixel((bch<<8)|bcl);
						}   
					}
				}
				temp+=(cfont.x_size/8);
			}
		}
	}
	else
	{
		temp=((c-cfont.offset)*((cfont.x_size/8)*cfont.y_size))+4;
		for(j=0;j<cfont.y_size;j++) 
		{
			for (zz=0; zz<(cfont.x_size/8); zz++)
			{
				ch=cfont.font[temp+zz];//ch=pgm_read_byte(&cfont.font[temp+zz]); 
				for(i=0;i<8;i++)
				{   
				
					if((ch&(1<<(7-i)))!=0)   
					{
						setXY(x+i+(zz*8),y+j,x+i+(zz*8)+1,y+j+1);
						setPixel((fch<<8)|fcl);
					} 
				}
			}
			temp+=(cfont.x_size/8);
		}
	}

	CS_DISEN;//sbi(P_CS, B_CS);
	clrXY();
}

void rotateChar(uint8_t c, int x, int y, int pos, int deg)
{
	uint8_t i,j,ch;
	uint16_t temp; 
	int newx,newy;
	double radian;
	radian=deg*0.0175; 
    int zz;

	CS_EN;//cbi(P_CS, B_CS);

	temp=((c-cfont.offset)*((cfont.x_size/8)*cfont.y_size))+4;
	for(j=0;j<cfont.y_size;j++) 
	{
		for (zz=0; zz<(cfont.x_size/8); zz++)
		{
			ch=cfont.font[temp+zz];//ch=pgm_read_byte(&cfont.font[temp+zz]);  
			for(i=0;i<8;i++)
			{   
				newx=x+(((i+(zz*8)+(pos*cfont.x_size))*cos(radian))-((j)*sin(radian)));
				newy=y+(((j)*cos(radian))+((i+(zz*8)+(pos*cfont.x_size))*sin(radian)));

				setXY(newx,newy,newx+1,newy+1);
				
				if((ch&(1<<(7-i)))!=0)   
				{
					setPixel((fch<<8)|fcl);
				} 
				else  
				{
					if (!_transparent)
						setPixel((bch<<8)|bcl);
				}   
			}
		}
		temp+=(cfont.x_size/8);
	}
	CS_DISEN;//sbi(P_CS, B_CS);
	clrXY();
}

void print(char *st, int x, int y, int deg)
{
	int stl, i;

	stl = strlen(st);

	if (orient==PORTRAIT)
	{
	if (x==RIGHT)
		x=(disp_x_size+1)-(stl*cfont.x_size);
	if (x==CENTER)
		x=((disp_x_size+1)-(stl*cfont.x_size))/2;
	}
	else
	{
	if (x==RIGHT)
		x=(disp_y_size+1)-(stl*cfont.x_size);
	if (x==CENTER)
		x=((disp_y_size+1)-(stl*cfont.x_size))/2;
	}

	for (i=0; i<stl; i++)
		if (deg==0)
			printChar(*st++, x + (i*(cfont.x_size)), y);
		else
			rotateChar(*st++, x, y, i, deg);
}

void printNumI(long num, int x, int y, int length, char filler)
{
	char buf[25];
	char st[27];
	bool neg=false;
	int c=0, f=0;
    int i;
  
	if (num==0)
	{
		if (length!=0)
		{
			for (c=0; c<(length-1); c++)
				st[c]=filler;
			st[c]=48;
			st[c+1]=0;
		}
		else
		{
			st[0]=48;
			st[1]=0;
		}
	}
	else
	{
		if (num<0)
		{
			neg=true;
			num=-num;
		}
	  
		while (num>0)
		{
			buf[c]=48+(num % 10);
			c++;
			num=(num-(num % 10))/10;
		}
		buf[c]=0;
	  
		if (neg)
		{
			st[0]=45;
		}
	  
		if (length>(c+neg))
		{
			for (i=0; i<(length-c-neg); i++)
			{
				st[i+neg]=filler;
				f++;
			}
		}

		for (i=0; i<c; i++)
		{
			st[i+neg+f]=buf[c-i-1];
		}
		st[c+neg+f]=0;

	}

	print(st,x,y,0);
}

void printNumF(double num, uint8_t dec, int x, int y, char divider, int length, char filler)
{
	char st[27];
	bool neg=false;
    int i;

	if (dec<1)
		dec=1;
	else if (dec>5)
		dec=5;

	if (num<0)
		neg = true;

	_convert_float(st, num, length, dec);

	if (divider != '.')
	{
		for (i=0; i<sizeof(st); i++)
			if (st[i]=='.')
				st[i]=divider;
	}

	if (filler != ' ')
	{
		if (neg)
		{
			st[0]='-';
			for (i=1; i<sizeof(st); i++)
				if ((st[i]==' ') || (st[i]=='-'))
					st[i]=filler;
		}
		else
		{
			for (i=0; i<sizeof(st); i++)
				if (st[i]==' ')
					st[i]=filler;
		}
	}

	print(st,x,y,0);
}

void setFont(uint8_t* font)
{
	cfont.font=font;
	cfont.x_size=fontbyte(0);
	cfont.y_size=fontbyte(1);
	cfont.offset=fontbyte(2);
	cfont.numchars=fontbyte(3);
}

uint8_t* getFont()
{
	return cfont.font;
}

uint8_t getFontXsize()
{
	return cfont.x_size;
}

uint8_t getFontYsize()
{
	return cfont.y_size;
}

void drawBitmap(int x, int y, int sx, int sy, bitmapdatatype data, int scale)
{
	unsigned int col;
	int tx, ty, tc, tsx, tsy;

	if (scale==1)
	{
		if (orient==PORTRAIT)
		{
			CS_EN;//cbi(P_CS, B_CS);
			setXY(x, y, x+sx-1, y+sy-1);
			for (tc=0; tc<(sx*sy); tc++)
			{
				col=data[tc];//col=pgm_read_word(&data[tc]);
				LCD_Write_BUS(col>>8,col & 0xff);
			}
			CS_DISEN;//sbi(P_CS, B_CS);
		}
		else
		{
			CS_EN;//cbi(P_CS, B_CS);
			for (ty=0; ty<sy; ty++)
			{
				setXY(x, y+ty, x+sx-1, y+ty);
				for (tx=sx-1; tx>=0; tx--)
				{
					col=data[(ty*sx)+tx];//col=pgm_read_word(&data[(ty*sx)+tx]);
					LCD_Write_BUS(col>>8,col & 0xff);
				}
			}
			CS_DISEN;//sbi(P_CS, B_CS);
		}
	}
	else
	{
		if (orient==PORTRAIT)
		{
			CS_EN;//cbi(P_CS, B_CS);
			for (ty=0; ty<sy; ty++)
			{
				setXY(x, y+(ty*scale), x+((sx*scale)-1), y+(ty*scale)+scale);
				for (tsy=0; tsy<scale; tsy++)
					for (tx=0; tx<sx; tx++)
					{
						col=data[(ty*sx)+tx];//col=pgm_read_word(&data[(ty*sx)+tx]);
						for (tsx=0; tsx<scale; tsx++)
							LCD_Write_BUS(col>>8,col & 0xff);
					}
			}
						CS_DISEN;//sbi(P_CS, B_CS);
		}
		else
		{
			CS_EN;//cbi(P_CS, B_CS);
			for (ty=0; ty<sy; ty++)
			{
				for (tsy=0; tsy<scale; tsy++)
				{
					setXY(x, y+(ty*scale)+tsy, x+((sx*scale)-1), y+(ty*scale)+tsy);
					for (tx=sx-1; tx>=0; tx--)
					{
						col=data[(ty*sx)+tx];//col=pgm_read_word(&data[(ty*sx)+tx]);
						for (tsx=0; tsx<scale; tsx++)
							LCD_Write_BUS(col>>8,col & 0xff);
					}
				}
			}
			CS_DISEN;//sbi(P_CS, B_CS);
		}
	}
	clrXY();
}

void drawBitmap_r(int x, int y, int sx, int sy, bitmapdatatype data, int deg, int rox, int roy)
{
	unsigned int col;
	int tx, ty, newx, newy;
	double radian;
	radian=deg*0.0175;  

	if (deg==0)
		drawBitmap(x, y, sx, sy, data, 1);
	else
	{
    	CS_EN;//cbi(P_CS, B_CS);
		for (ty=0; ty<sy; ty++)
			for (tx=0; tx<sx; tx++)
			{
				col=data[(ty*sx)+tx];//col=pgm_read_word(&data[(ty*sx)+tx]);

				newx=x+rox+(((tx-rox)*cos(radian))-((ty-roy)*sin(radian)));
				newy=y+roy+(((ty-roy)*cos(radian))+((tx-rox)*sin(radian)));

				setXY(newx, newy, newx, newy);
				LCD_Write_BUS(col>>8,col & 0xff);
			}
		CS_DISEN;//sbi(P_CS, B_CS);
	}
	clrXY();
}

int getDisplayXSize()
{
	if (orient==PORTRAIT)
		return disp_x_size+1;
	else
		return disp_y_size+1;
}

int getDisplayYSize()
{
	if (orient==PORTRAIT)
		return disp_y_size+1;
	else
		return disp_x_size+1;
}


//************************************************************

void DelayMs(unsigned short int ms) {
    unsigned char i;
    while (ms--) {
        i = 4;
        while (i--) {
            Delay10us(25);
        }
    }
}

void Delay10us(unsigned long dwCount) {
    volatile unsigned long _dcnt;

    _dcnt = dwCount * ((unsigned long) (0.00001 / (1.0 / SYS_TIME_CPU_CLOCK_FREQUENCY) / 10));
    while (_dcnt--) {
        Nop();
        Nop();
        Nop();
    }
}

//*******************************************************************************