/*******************************************************************************
* Copyright (C) 2020 Microchip Technology Inc. and its subsidiaries.
*
* Subject to your compliance with these terms, you may use Microchip software
* and any derivatives exclusively with Microchip products. It is your
* responsibility to comply with third party license terms applicable to your
* use of third party software (including open source software) that may
* accompany Microchip software.
*
* THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
* EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
* WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
* PARTICULAR PURPOSE.
*
* IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT, SPECIAL, PUNITIVE,
* INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE OF ANY KIND
* WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF MICROCHIP HAS
* BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE. TO THE
* FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN
* ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY,
* THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
*******************************************************************************/

/*******************************************************************************
  ILI9488 Display Top-Level Driver Source File

  File Name:
    drv_gfx_ili9488.c

  Summary:
    Top level driver for ILI9488.

  Description:
    Build-time generated implementation for the ILI9488 Driver.

    Created with MPLAB Harmony Version 3.0
*******************************************************************************/




#include "definitions.h"

#include "drv_gfx_ili9488_cmd_defs.h"
#include "drv_gfx_ili9488.h"
//#include "gfx/interface/drv_gfx_disp_intf.h"

#include "../../UTFT.h"
#include "../../DefaultFonts.c"

// Default max width/height of ILI9488 frame
#define DISPLAY_DEFAULT_WIDTH   320
#define DISPLAY_DEFAULT_HEIGHT  240

#define DISPLAY_WIDTH   320
#define DISPLAY_HEIGHT  240

#define PIXEL_BUFFER_COLOR_MODE GFX_COLOR_MODE_RGB_565

#define SCREEN_WIDTH DISPLAY_WIDTH
#define SCREEN_HEIGHT DISPLAY_HEIGHT


/** ILI9488_DRV_STATE

  Summary:
    Enum of ILI9488 driver states.
    
*/
typedef enum
{
    INIT = 0,
    IDLE,
    BLIT_COLUMN_CMD,
    BLIT_COLUMN_DATA,
    BLIT_PAGE_CMD,
    BLIT_PAGE_DATA,
    BLIT_WRITE_CMD,
    BLIT_WRITE_DATA,
    BLIT_WAIT,
    ERROR,
} ILI9488_DRV_STATE;

/** ILI9488_DRV

  Summary:
    Structure contains driver-specific data and ops pointers.
    
*/
typedef struct ILI9488_DRV 
{   
    /* Driver state */
    ILI9488_DRV_STATE state;

    /* Port-specific private data */
    void *port_priv;

    struct
    {
        int32_t x;
        int32_t y;
        gfxPixelBuffer* buf;
    } blitParms;

} ILI9488_DRV;

//static uint16_t pixelBuffer[SCREEN_WIDTH];
static ILI9488_DRV drv;
static uint32_t swapCount = 0;


/* ************************************************************************** */

/**
  Function:
    static int ILI9488_Reset(void)

  Summary:
    Toggles the hardware reset to the ILI9488.

  Description:
    This function toggles the GPIO pin for asserting reset to the ILI9488.

  Parameters:
    None

  Returns:
    None

*/

gfxResult DRV_ILI9488_Initialize(void)
{
    drv.state = INIT;

    return GFX_SUCCESS;
}

void DRV_ILI9488_Update(void)
{
    uint16_t* ptr;
    
    switch (drv.state)
    {
        case INIT:
        {        
            // perform driver initialization here
            InitLCD(LANDSCAPE);
            setColor_w(VGA_YELLOW);
            setBackColor(0, 40, 0);
            setFont(BigFont);         
            
            drv.state = IDLE;

            break;
        }
        
        case BLIT_COLUMN_CMD:
        {
  
            drv.state = IDLE;

            ptr = gfxPixelBufferOffsetGet_Unsafe(drv.blitParms.buf, 0, 0);

            drawBitmap(drv.blitParms.x, drv.blitParms.y,
                      drv.blitParms.buf->size.width, drv.blitParms.buf->size.height, ptr, 1);
                
            break;
             
        }
        case IDLE:
        case ERROR:            
        default:
        {
            break;
        }        
    }
}

gfxResult DRV_ILI9488_BlitBuffer(int32_t x,
                                 int32_t y,
                                 gfxPixelBuffer* buf)
{
    if(drv.state != IDLE)
        return GFX_FAILURE;

    drv.blitParms.x = x;
    drv.blitParms.y = y;
    drv.blitParms.buf = buf;
    drv.state = BLIT_COLUMN_CMD;

    while (drv.state != IDLE)
    {
        DRV_ILI9488_Update();
    }
    
    return GFX_SUCCESS;
}

gfxDriverIOCTLResponse DRV_ILI9488_IOCTL(gfxDriverIOCTLRequest req,
                                         void* arg)
{
    gfxIOCTLArg_Value* val;
    gfxIOCTLArg_DisplaySize *disp;
    gfxIOCTLArg_LayerRect *rect;

	switch(req)
	{
		case GFX_IOCTL_GET_COLOR_MODE:
		{
			val = (gfxIOCTLArg_Value*)arg;
			
			val->value.v_uint = PIXEL_BUFFER_COLOR_MODE;
			
			return GFX_IOCTL_OK;
		}
		case GFX_IOCTL_GET_BUFFER_COUNT:
		{
			val = (gfxIOCTLArg_Value*)arg;
			
			val->value.v_uint = 1;
			
			return GFX_IOCTL_OK;
		}
		case GFX_IOCTL_GET_DISPLAY_SIZE:
		{
			disp = (gfxIOCTLArg_DisplaySize*)arg;			
			
			disp->width = SCREEN_WIDTH;
			disp->height = SCREEN_HEIGHT;
			
			return GFX_IOCTL_OK;
		}
		case GFX_IOCTL_GET_LAYER_COUNT:
		{
			val = (gfxIOCTLArg_Value*)arg;
			
			val->value.v_uint = 1;
			
			return GFX_IOCTL_OK;
		}
		case GFX_IOCTL_GET_ACTIVE_LAYER:
		{
			val = (gfxIOCTLArg_Value*)arg;
			
			val->value.v_uint = 0;
			
			return GFX_IOCTL_OK;
		}
		case GFX_IOCTL_GET_LAYER_RECT:
		{
			rect = (gfxIOCTLArg_LayerRect*)arg;
			
			rect->x = 0;
			rect->y = 0;
			rect->width = SCREEN_WIDTH;
			rect->height = SCREEN_HEIGHT;
			
			return GFX_IOCTL_OK;
		}
		case GFX_IOCTL_GET_VSYNC_COUNT:
		{
			val = (gfxIOCTLArg_Value*)arg;
			
			val->value.v_uint = swapCount;
			
			return GFX_IOCTL_OK;
		}
		case GFX_IOCTL_LAYER_SWAP:
		{
			swapCount += 1;
			
			return GFX_IOCTL_OK;
		}
		default:
		{
		}
	}
	
	return GFX_IOCTL_UNSUPPORTED;
}
